//Date:  March 12, 2020
//Author:  Matt Meerian
//compiler-IDE:  Arduino 1.8.2 IDE
//Target MCU:  ATMEGA328PB, 16MHz  (the board in the arduino IDE is set to the "nano")
//  NOTE:  The 'PB version of the MCU adds a couple of timers and another UART.  (none of the new features are used in this software)
//Description:  A rotary encoder allows the user to make menu selections on a 128x32 OLED display.
//  The menu selections control the PWM going to a white LED PCB(s).  The system works only on 12V LED boards.
//  A switchmode power supply converts 12VDC to 5VDC for the MCU.  (on rev B of the board, it is a linear regulator (SOT223 package))
//  MCU notes:  
//    -Timer 0 is used by the arduino IDE
//    -Timesr 1 generates the PWM for the output and also the overflow generates a roughly 20mS "tock" for the main routine (no in-line delays!)
//    -The PWM output on the PCB was originally put on the timer 0 PWM output, but a mod wire was placed to have timer 1 set the PWM, on Rev A of the PCB

/*
  version history:
    2020-5-11:  V1.2
      -increased the PWM frequency from 122Hz to 975Hz.
      -When the "gaming" dim area is entered in the menu, the output immediately changes to the dim level
      -there are dimming level steps over a greater range
      -added a define statement for the maximum PWM setting (for the tenergy lamp)
    2020-5-14:  V1.3
      -When the display goes to sleep, the menu goes back to the starting postion
*/

#define ENCODER_OPTIMIZE_INTERRUPTS //use hardware interrupts, It must be defined before Encoder.h is included.

#include <Arduino.h>
#include <Wire.h>   //used for the I2C interface for the 128x32 pixel OLED
#include <EEPROM.h>

//You must install these libraries into the arduino IDE:
#include <Encoder.h>  //Rotary encoder, https://www.pjrc.com/teensy/td_libs_Encoder.html
#include <U8g2lib.h>  //OLED graphics, https://github.com/olikraus/u8g2/wiki
//The board in Arduino should be set as "ATMEGA329PB crystal clock" using this board library:
//  https://github.com/watterott/ATmega328PB-Testing
//  Note:  When this is done, it breaks the "encoder" library.  You need to go into "interrupt_pins.h" under "C:\Users\mmeerian\Documents\Arduino\libraries\Encoder\utility"
//  and then go to roughly line 35 and add " || defined(__AVR_ATmega328PB__) " to the end.

Encoder myEnc(2, 3);  //use pins with the external interrupts available
U8G2_SSD1306_128X32_UNIVISION_F_HW_I2C u8g2(U8G2_R0, /* reset=*/ U8X8_PIN_NONE);  // Adafruit ESP8266/32u4/ARM Boards + FeatherWing OLED

#define NUM_MENU_ELEMENTS 5  //stores the maximum number of menu elements
#define START_UP_NUM_ELEMENTS 2   //we have three possible selections
#define ENTER_PIN A3  //The rotary encoder has a switch that closes when you press in on the shaft, this pin goes low when this happens
#define debug     //when commentted in, it enables debugging on the OLED
#define MAX_PWM 700   //defines the max PWM, for timer 1, that is 0x3FF.  limit to 700 for the tenergy lamp

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#define MINUTE_RELOAD 14520    //counting down to a minute, 60sec/4.132mS = 14520
#define EEPROM_ADDR_GAMING  1 //EEPROM address for storing user variables
#define EEPROM_ADDR_LED     2 //EEPROM address for storing user variables
#define EEPROM_ADDR_DISPLAY 3 //EEPROM address for storing user variables
#define EEPROM_ADDR_START_UP  4   //EEPROM address for storing what the LED's should be doing at startup
#define EEPROM_ADDR_PREVIOUS  5   //when the startup is set to "previous", this stores that value
#define EEPROM_ADDR_HEARTBEAT  6   //when 1, the heartbeat LED blinks; when 0, the hearbeat LED does not blink

unsigned char u08MenuSelection=0;   //start at the first position in the menu
unsigned char u08OldMenuSelection=0;  //this value tracks the last menu postion
unsigned char u08TmrEncoder=4;  //the encoder has 4 pulses per click, this counts down to 4 pulses
unsigned int counting=0;  //debug:  used for counting the number of button presses
unsigned char u08EnterReleased=0;  //allows the enter button to run code only once, when it is first pressed
long newPosition; //keeps track of the encoder postion
long oldPosition  = 0;   //stores the encoders old position, was -999
unsigned char u08State=0; //when state is 0, we are in the upper menu; when 1, we are in the user value adjustment area
bool boolUpdateDisplay=true;   //when true, the display needs to be updated
signed int s16LowerMovement=0;  //a "user adjustment" value for testing
unsigned int u16OutputLevel=0; //controls the intensity of the main output LED, 0 is off, 0x3F(1024) is full on
unsigned char u08StartUpState=0;  //for the 5th element in the menu, on startup, should the LED's be on, off, or dim
unsigned char u08Heartbeat=0;   //when "1", the heartbeat indicator LED is active; when "0", the heartbeat indicator LED is not active.

//static variables are in ISR's
static unsigned char u08Tock; //keeps a 20.66mS second "tock"
static unsigned int u16TmrMinute=MINUTE_RELOAD;
static bool boolProcessAMinute=false;
static unsigned char u08Heart;

unsigned int u16Count=0;

const char strMenuText_0[] PROGMEM = {"FULL ON/OFF"};
const char strMenuText_1[] PROGMEM = {"GAMING"};
const char strMenuText_2[] PROGMEM = {"LED SLEEP"};
const char strMenuText_3[] PROGMEM = {"DISPLAY SLEEP"}; 
const char strMenuText_4[] PROGMEM = {"ON POWER UP"}; 
const char strMenuText_5[] PROGMEM = {"HEARTBEAT"}; 
const char strMenuText_6[] PROGMEM = {"END 2"};
const char strMenuText_7[] PROGMEM = {"END 3"};

const char strUserText_0[] PROGMEM = {"Output: "};
const char strUserText_1[] PROGMEM = {"Dim Level(%): "};
const char strUserText_2[] PROGMEM = {"Turn Off(Hrs): "};
const char strUserText_3[] PROGMEM = {"Screen Saver: "};
const char strUserText_4[] PROGMEM = {"LED's are: "};
const char strUserText_5[] PROGMEM = {"Indicator is: "};
const char strUserText_6[] PROGMEM = {"End B"};
const char strUserText_7[] PROGMEM = {"End C"};

const char str3rdLine_0[] PROGMEM = {"Long Press=Dim"};
const char str3rdLine_1[] PROGMEM = {""};
const char str3rdLine_2[] PROGMEM = {"Left(min): "};
const char str3rdLine_3[] PROGMEM = {"sleep(min): "};
const char str3rdLine_4[] PROGMEM = {""};
const char str3rdLine_5[] PROGMEM = {""};
const char str3rdLine_6[] PROGMEM = {""};
const char str3rdLine_7[] PROGMEM = {""};

const char strStartUp_0[] PROGMEM = {"ON"};
const char strStartUp_1[] PROGMEM = {"OFF"};
const char strStartUp_2[] PROGMEM = {"PREV"};

const PROGMEM unsigned char dimLevel[] = {1,2,5,10,15,20,30,40,50,60,70,80,90,55};  //only 10 positions used for dimming
#define DIM_NUM_ELEMENTS 9
const PROGMEM unsigned char ledSleepLevel[] = {1,2,3,6,9,12,24,200};
#define LED_NUM_ELEMENTS 7
const PROGMEM unsigned char displaySleepLevel[] = {1,2,5,10,15,30,60,90};
#define DISPLAY_NUM_ELEMENTS 6
const PROGMEM unsigned char powerUpLevel[] = {1,2,3};
#define POWER_NUM_ELEMENTS 2

const char *const menu_table[] PROGMEM = {strMenuText_0, strMenuText_1, strMenuText_2, strMenuText_3, strMenuText_4, strMenuText_5, strMenuText_6, strMenuText_7};
const char *const user_table[] PROGMEM = {strUserText_0, strUserText_1, strUserText_2, strUserText_3, strUserText_4, strUserText_5, strUserText_6, strUserText_7};
const char *const user_3rdLine[] PROGMEM = {str3rdLine_0, str3rdLine_1, str3rdLine_2, str3rdLine_3, str3rdLine_4, str3rdLine_5, str3rdLine_6, str3rdLine_7};
const char *const user_StartUp[] PROGMEM = {strStartUp_0, strStartUp_1, strStartUp_2};

char buft[20];  //RAM buffer to hold strings from flash for the menu
unsigned char u08MenuLength;
bool boolOutputOn=true;
unsigned char u08DimLevel=0;
unsigned char u08DimTarget=5;
unsigned char u08LedSleepLevel=0;
unsigned char u08LedSleepTarget=2;
unsigned char u08DisplaySleepLevel=0;
unsigned char u08DisplaySleepTarget=5;
unsigned char u08PowerUpLevel=0;
unsigned char u08PowerUpTarget=1;
unsigned char u08CntLongEnterPress=0;
unsigned int u16CntLedSleep=0;
unsigned char u08CntDisplaySleep=0;

//When we are in the "user" area, this updates the values to be displayed on the screen.  
void updateUserValues(void)
{
  unsigned char fullon=0;
 
  switch (u08MenuSelection) 
  {
    case 0:  
      if(u16OutputLevel==0)
      {
        u8g2.print("OFF");    //should be in flash
      }
      else
      {
        u8g2.print("ON");   //should be in flash
      }
    break;
    case 1:  
      u8g2.print(u08DimTarget); //show the user value changing        
    break;
    case 2:  
      u8g2.print(u08LedSleepTarget); //show the user value changing        
    break;
    case 3:  
      u8g2.print(u08DisplaySleepTarget); //show the user value changing        
    break;
    case 4:
      if(u08StartUpState==0)
      {
        u8g2.print("ON");    //should be in flash
      }
      else if(u08StartUpState==1)
      {
        u8g2.print("OFF");   //should be in flash
      }
      else
      {
        u8g2.print("DIM");   //should be in flash
      }
    break;
    case 5:
      if(u08Heartbeat==0)
      {
        u8g2.print("OFF");    //should be in flash
      }
      else
      {
        u8g2.print("ON");   //should be in flash
      }
    break;
    default:  
      u8g2.print("Err"); //should be in flash, show an error
  }
}

//get the stored values out of EEPROM when the MCU is booting up
void readEepromOnStartup(void)
{
  u08DimLevel = EEPROM.read(EEPROM_ADDR_GAMING);  //get the user value out of EEPROM at power up
  if(u08DimLevel>DIM_NUM_ELEMENTS)  //check to make sure the value is valid
  {
    u08DimLevel=0;
    EEPROM.write(EEPROM_ADDR_GAMING, 0);    //The EEPROM value must have not been good, write a valid value
  }
  u08DimTarget=pgm_read_byte(&(dimLevel[u08DimLevel]));             

  u08LedSleepLevel= EEPROM.read(EEPROM_ADDR_LED);  //get the user value out of EEPROM at power up
  if(u08LedSleepLevel>LED_NUM_ELEMENTS)  //check to make sure the value is valid
  {
    u08LedSleepLevel=0;
    EEPROM.write(EEPROM_ADDR_LED, 0);    //The EEPROM value must have not been good, write a valid value
  }
  u08LedSleepTarget=pgm_read_byte(&(ledSleepLevel[u08LedSleepLevel]));             
  u16CntLedSleep=u08LedSleepTarget*60;  //put the number of hours for the LED's to go to sleep into minutes.

  u08DisplaySleepLevel= EEPROM.read(EEPROM_ADDR_DISPLAY);  //get the user value out of EEPROM at power up
  if(u08DisplaySleepLevel>DISPLAY_NUM_ELEMENTS)  //check to make sure the value is valid
  {
    u08DisplaySleepLevel=0;
    EEPROM.write(EEPROM_ADDR_DISPLAY, 0);    //The EEPROM value must have not been good, write a valid value
  }
  u08DisplaySleepTarget=pgm_read_byte(&(displaySleepLevel[u08DisplaySleepLevel]));
  u08CntDisplaySleep=u08DisplaySleepTarget;   //start the count down to turning off the display

  u08StartUpState= EEPROM.read(EEPROM_ADDR_START_UP);  //get the user value out of EEPROM at power up
  if(u08StartUpState>START_UP_NUM_ELEMENTS)  //check to make sure the value is valid
  {
    u08StartUpState=0;
    EEPROM.write(EEPROM_ADDR_START_UP, 0);    //The EEPROM value must have not been good, write a valid value
  }

  u08Heartbeat= EEPROM.read(EEPROM_ADDR_HEARTBEAT);  //get the user value out of EEPROM at power up
  if(u08Heartbeat>1)  //check to make sure the value is valid
  {
    u08Heartbeat=0;
    EEPROM.write(EEPROM_ADDR_HEARTBEAT, 0);    //The EEPROM value must have not been good, write a valid value
  }
}

//this routine runs every minute
void processMinute(void)
{
  if(u16CntLedSleep>0)  //we are counting down every minute to the LED sleep?
  {
    u16CntLedSleep--; //yes, so count it down  (any user button press will reset the count)
  }
  else  //we have counted down, now the LED's need to shut off
  {
    u16OutputLevel=0;   //turn off the output LED's
    u16CntLedSleep=u08LedSleepTarget*60;  //start up the timer again (not really needed)
  }
  if(u08CntDisplaySleep>0)  //we are counting down every minute to the display sleeping
  {
    u08CntDisplaySleep--;   //counting down to turning off the display
    Serial.print(F("cntdisplaysleep:  "));
    Serial.println(u08CntDisplaySleep);
    boolUpdateDisplay=true;   //allow the display to be updated to show when the OLED will turn off 
  }
  else  //it is time to put the display to sleep (well, at least clear the display)
  {
    //u08CntDisplaySleep=u08DisplaySleepTarget;  //reload the timer, not really needed, any button press will light up the display
    u08MenuSelection=0;   //go back to the left most menu selection (so we always start at the same place when the display comes out of sleep)
    u08OldMenuSelection=0;  //we do not want the display to update at this moment (the user did not press enter or rotate the knob)
    u08State=0;   //make sure the left/right menu selection is operational
    Serial.print(F("display sleeping:  "));
    Serial.println(u08State);
    u8g2.clearBuffer(); //clear the contents in ram, any button press will turn the display back on
    u8g2.sendBuffer();  //transfer internal memory to the display  NOTE:  must be used
  }
}

void setup() {
  PORTB=B11111000;       //to enable a pullup resistor on this port, set the bit to "1"
  DDRB = DDRB | B00000111;  //set pins to input or output
  PORTD=B00000000;       //to enable a pullup resistor on this port, set the bit to "1"
  DDRD = DDRD | B10110011;  //set pins to input or output

  //delay(2000);  //   !!!!!!Fuking read this --->  NOTE:  on powerup, the "DIY MORE", 128x32 OLED display needs as much as
  //  2 seconds to stablize before it can be initialized.  Most the white text OLED display does not need this.
  
  u8g2.begin();   //start the u8g2
  Serial.begin(115200);   //serial port initialization
  Serial.println(F("Hello"));
//  Serial.println(F("Basic Encoder Test:"));
  u8g2.clearBuffer(); //clear the contents in ram, any button press will turn the display back on
  u8g2.setFont(u8g2_font_ncenB08_tr); // choose a suitable font  (font is 7 wide and 9 tall, this includes one pixel space)
  u8g2.setCursor(50, 20);
  u8g2.print("V1.3");   //should be in flash
  u8g2.setCursor(40, 30);
  u8g2.print("Tenergy");   //should be in flash
  u8g2.sendBuffer();  //transfer internal memory to the display  NOTE:  must be used
  delay(500);
  


  pinMode(ENTER_PIN, INPUT_PULLUP);
  cli();//stop interrupts while we set up the timer

  //Timer 0 is used for the arduino "delay" function and other things.
  
  //timer 1 setup:  16 bit timer, controls the PWM output going to the mosfet
  TCCR1B = B00000000;// Stop Timer/Counter1 clock by setting the clock source to none.
  TCCR1A = B00000000;// Set Timer/Counter1 to normal mode.
  TCNT1  = 0;// Set Timer/Counter1 to 0
  OCR1A = 30;// Set the Output Compare A for Timer/Counter1, 100 Hz
  //TCCR1A = B10000011;// Set Timer/Counter1 to PWM phase correct, 0x3FF (0d1024) maximum value
  TCCR1A |= (1<<COM1A1)|(1<<WGM11)|(1<<WGM10);// Set Timer/Counter1 to PWM phase correct, 0x3FF (0d1024) maximum value
  TCCR1B |= (1<<CS11);  //|(1<<CS10);   //prescaler set to divide by 8

  //NOTE:  timer 2 is only on the "PB" version of the ATMEGA328PB, this is an 8-bit timer.  
  //  An overflow is ued for the roughly 20mS "tock"
  TCCR2B = B00000000;// Stop Timer/Counter2 clock by setting the clock source to none.
  TCCR2A = B00000000;// Set Timer/Counter2 to normal mode.
  TCNT2  = 0;// Set Timer/Counter1 to 2
  //OCR2A = 20;// Set the Output Compare A for Timer/Counter1, 100 Hz
  TCCR2B |= (1<<CS22)|(1<<CS21); // Start Timer/Counter2 clock by setting the source to CPU source. Set prescalar to 1/8 (2Mhz).
  TIMSK2 =B00000001;  //enable timer 2 overflow interrupt
  
  sei();//allow interrupts
  readEepromOnStartup();    //get the user values out of EEPROM and check to see if they are valid.  If they are not valid, then correct the values.
  if(u08StartUpState==0) //The LED's should be "ON" based on the EEPROM value
  {
    u16OutputLevel=MAX_PWM;
  }
  else if (u08StartUpState==1)  //The LED's should be "OFF" based on the EERPOM value
  {
    u16OutputLevel=0;
  }
  else if(u08StartUpState==2)    //The LED's should be "DIM" based on the EERPOM value
  {
    u16OutputLevel=u08DimTarget;
  }
  else  //previous state
  {
    
  }

  
  
}

//Timer 2 (8 bit), overflow interrupt,  used as a timer "tock", prevents in-line delays, sleep() is evil
ISR(TIMER2_OVF_vect)          // timer compare interrupt service routine
{
  if(u08Tock>0)
  {
    u08Tock--;  //counting down to a tock, the main routine reloads the tock
  }
  if(u16TmrMinute>0)  //counting down to a minute
  {
    u16TmrMinute--;
  }
  else  //a minute has passed!
  {
    u16TmrMinute=MINUTE_RELOAD; //reload the minute timer.
    boolProcessAMinute=true;
  }

  if(u08Heartbeat==1)
  {
    u08Heart--;   //counting down to toggling the on board, red LED
    if(u08Heart==0)
    {
      digitalWrite(5, !digitalRead(5));    //toggle PortD, pin5, the red on-board LED
      u08Heart=60;  //reset the count down timer.
    }  
    //digitalWrite(10, !digitalRead(10));    //toggle PortB pin 2, ATMEGA328 pin 14, arduino pin 10 (note:  you must set pin 10 as an output
  }
  else  //make sure the LED is off if the indicator is disabled
  {
    digitalWrite(5,LOW);    //PortD, pin5, the red on-board LED
  }
}

//With a 16MHz clock and the TCCR1B register settings, this routine will run every 4.132mS.
//Not used anymore
ISR(TIMER1_OVF_vect)          // timer compare interrupt service routine
{

}

void loop() {
  if(u08Tock==0)  //only read the buttons/update the display every 20.66mS
  {
    u08Tock=5;  //This part will run every 4.132mS * 5 = 20.66mS
    //digitalWrite(10, !digitalRead(10));    //toggle PortB pin 2, ATMEGA328 pin 14, arduino pin 10 (note:  you must set pin 10 as an output)
    
    if((u16OutputLevel*10)<0x3FF)  //must be kept lower than 10 bits for the PWM
    {
      OCR1A=u16OutputLevel*10;   //sets the PWM going to the LED string, uses timer1  
    }
    else
    {
      OCR1A=MAX_PWM;  //there must have been an error, so set the PWM to the maximum allowable value 
    }
    if(boolProcessAMinute==true)  //a minute has passed!   So, process the timing
    {
      processMinute();
      boolProcessAMinute=false;   //reset the minute timer, the ISR counts this down
    }

  //----------------------------------------------------------------------------------------------------------------------------------------------
  //read the enter pin (pushing in on the rotary encoder)
    int reading = digitalRead(ENTER_PIN); //get the "enter" button presses, only when it is first pressed
    if(reading==0)  //is the button pressed down?
    {
      u08CntDisplaySleep=u08DisplaySleepTarget;   //start the newcount down to turning off the display
      if(u08EnterReleased==1)  //it must be the first time it has been pressed in awhile
      {
        u08EnterReleased=0;  //don't go in here again, unless the button is released    
        //Serial.print(F("Pressed  "));  //debug
        //Serial.println(counting); //debug
        if(u08MenuSelection==0) //if we are on the first menu page, any button press turns on/off the output LED's
        {
          if(u16OutputLevel>0)  //are the LED's on?
          {
            u16OutputLevel=0;  //then turn them off
          }
          else  //the LED's must be off
          {
            u16OutputLevel=MAX_PWM;   //turn on the output LED's
          }
        }
        if(u08MenuSelection==1) //if we are entering into the "gaming" dim setting, then set the PWM to the displayed value
        {
          u16OutputLevel=u08DimTarget;  
          Serial.println(u08DimTarget);  
        }
        if(u08State==0) //are we in the menus?
        {
          if(u08MenuSelection!=0) //do not go into the user area if we are on the first menu page, this should just turn on/off the main LED's
          {
            u08State=1; //then move to the user input area
          }
        }
        else  //we must be in the user input area
        {
          u08State=0; //go back to the menus
          switch (u08MenuSelection)   //we need to save the user value to EEPROM, when the user is exiting the user entry area
          {
            case 0:  
              //no need to save anything to EEPROM here
            break;
            case 1:  
              EEPROM.write(EEPROM_ADDR_GAMING, u08DimLevel);    //Save the EEPROM value (even if it was not changed by the user)
            break;
            case 2:  
              EEPROM.write(EEPROM_ADDR_LED, u08LedSleepLevel);    //Save the EEPROM value (even if it was not changed by the user)
              u08LedSleepTarget=pgm_read_byte(&(ledSleepLevel[u08LedSleepLevel]));             
              u16CntLedSleep=u08LedSleepTarget*60;  //reset the timer, since the user made a new selection
            break;
            case 3:  
              EEPROM.write(EEPROM_ADDR_DISPLAY, u08DisplaySleepLevel);    //Save the EEPROM value (even if it was not changed by the user)
              u08DisplaySleepTarget=pgm_read_byte(&(displaySleepLevel[u08DisplaySleepLevel]));
              u08CntDisplaySleep=u08DisplaySleepTarget;   //start the newcount down to turning off the display
            break;
            case 4:
              EEPROM.write(EEPROM_ADDR_START_UP, u08StartUpState);
              //Serial.println(F("u08StartUp d= "));
              //Serial.println(u08StartUpState);
            break;
            case 5:
              EEPROM.write(EEPROM_ADDR_HEARTBEAT, u08Heartbeat);
            break;
          }          
        }
        boolUpdateDisplay=true; //have the display updated
        if(counting<10000)  //debug code, not needed
        {
            counting++; //debug, used for counting the number of enter button presses
        }
      }
      
      if(u08CntLongEnterPress<50) //is the enter button being held down, anywhere in the menu?
      {
        u08CntLongEnterPress++;   //counting up to going into gaming mode
      }
      else
      {
        u16OutputLevel=pgm_read_byte(&(dimLevel[u08DimLevel]));  //There has been a long press on the enter button, set the output led's at their gaming dim level.
      }
    }
    else
    {
      u08EnterReleased=1;  //the enter button is not being held down
      u08CntLongEnterPress=0;   //the enter button has been released, reset the counter, it only goes up when being held down.
    }

    
  //----------------------------------------------------------------------------------------------------------------------------------------------
  //update the display
  
    if(boolUpdateDisplay==true)   //has a flag been set to update the display?
    {
      boolUpdateDisplay=false;  //We are going to update the display, so reset the flag for the next update
      u8g2.clearBuffer();          // clear the internal memory
      u8g2.setFont(u8g2_font_ncenB08_tr); // choose a suitable font  (font is 7 wide and 9 tall, this includes one pixel space)
      u8g2.setDrawColor(1);   //set the color of text back to normal
      strcpy_P(buft, (char *)pgm_read_word(&(menu_table[u08MenuSelection]))); //get the correct menu information out of flash and save it into a ram buffer
      u08MenuLength=u8g2.getStrWidth(buft);   //find the length of the string from flash
      if(u08State==0) //Is the menu active?,  write the top line on the OLED
      {
        u8g2.drawBox(62-(u08MenuLength/2), 0, u08MenuLength+4, 10); //draw the inverted color background (highlight) box, 64 is the center, 62 adds to pixels on the left side.  The +4 for the box length adds two pixels on the other side
        u8g2.setDrawColor(0);     //invert the colors
        u8g2.setCursor(64-(u08MenuLength/2), 9);  //set the start area of the text so it is centered on the screen
        u8g2.print(buft);   //write the menu text to the OLED
        u8g2.setDrawColor(1);   //set the color of text back to normal
      }
      else  //The user input area must be active (the menu is not active), write the top line on the OLED
      {
        u8g2.setCursor(64-((u08MenuLength)/2), 9);
        u8g2.print(buft);
      }

      strcpy_P(buft, (char *)pgm_read_word(&(user_table[u08MenuSelection])));    //get the correct user information out of flash and save it into a ram buffer
      u08MenuLength=u8g2.getStrWidth(buft);   //find the length of the string from flash
      if(u08State==0) //is the menu area active?, write the second line on the OLED
      {
        u8g2.setCursor(0, 20);
        u8g2.print(buft);
        updateUserValues();
      }
      else  //the user input area must be active, write the second line 
      {
        u8g2.drawBox(u08MenuLength-2, 10, 45, 10);    //draw a box for the background, first X and Y is the upper left, second X and Y is the height and width, the -2 slides the box to the left a little
        u8g2.setCursor(0, 20);  //For the text on the second line, set it to the left side
        u8g2.print(buft);   //write the text in the RAM buffer
        u8g2.setDrawColor(0); //invert the text colors
        updateUserValues();
        u8g2.setDrawColor(1); //set the colors back to normal
      }

#ifdef debug  //show debugging information in the third line of the OLED
      strcpy_P(buft, (char *)pgm_read_word(&(user_3rdLine[u08MenuSelection])));    //get the correct user information out of flash and save it into a ram buffer
      u08MenuLength=u8g2.getStrWidth(buft);   //find the length of the string from flash
      u8g2.setCursor(0, 30);  //set the cursor position to the third line
      u8g2.print(buft);   //print the flash information from the ram buffer

      if(u08MenuSelection==2) //debug, show the minutes counting down to LED sleep
      {
        u8g2.print(u16CntLedSleep);        //does not update in real time
      }
      if(u08MenuSelection==3)   //debug, show the minutes counting down to OLED display sleep
      {
        u8g2.print(u08CntDisplaySleep);   //show the amount of time left to putting the OLED to sleep
      }
#endif

      u8g2.setCursor(104, 30);  //set the location of the page number (i.e. 3 of 6) pages
      u8g2.print(u08MenuSelection+1);   //put up the current page we are on, it actually starts at zero, but we want 1 to 6
      u8g2.print("of6");      //should be put into flash
      
      if(u08MenuSelection!= NUM_MENU_ELEMENTS)  //don't show the right arrow when we are at the end of the menu
      {
        u8g2.drawTriangle(118,0, 118,10, 123,5);  //draw the right arrow
      }
      if(u08MenuSelection!=0)  //don't show the left arrow when we are at the end of the menu
      {
        u8g2.drawTriangle(5,5, 10,0, 10,10);  //draw the left arrow
      }
      u8g2.sendBuffer();          //transfer internal memory to the display  NOTE:  must be used
    }
    if(u08MenuSelection!=u08OldMenuSelection) //has the menu selection changed from last time through the loop?
    {
        u08OldMenuSelection=u08MenuSelection; //yes, it has, so update the current menu selection
        boolUpdateDisplay=true;   //refresh the display
    }
    
  //----------------------------------------------------------------------------------------------------------------------------------------------
  //read the encoder
  //encoder processing
      newPosition = myEnc.read(); //read the rotary encoder, each "click", generates about 4 pulses
      if (newPosition != oldPosition && u08TmrEncoder==0) //look for a postion change in the encoder and that we have finished processing the previous click
      { //NOTE:  this lets the encoder settle before the information is processed
        u08TmrEncoder=10;   //load up the timer to process the encoder click,   ~20mS * 9 = 180mS, the user doesn't really notice this
      }
      if(u08TmrEncoder>0) //if the rotary encoder had moved, then we are counting down to processing
      {
        u08TmrEncoder--;  
      }
      if(u08TmrEncoder==1)  //It is time to process the rotary click!
      {
          u08CntDisplaySleep=u08DisplaySleepTarget;   //start the newcount down to turning off the display
          updateDisplayElements();
          oldPosition = newPosition;  //store the new position of the encoder
      }
  }
  //----------------------------------------------------------------------------------------------------------------------------------------------
}

//This routine controls what happens when the rotary encoder is rotated left or right
void updateDisplayElements(void)
{
  if(newPosition>oldPosition)   //are we turning clockwise?
  {
    if(u08State==0) //only allow the menu to change if we are in that area
    {
      if(u08MenuSelection<NUM_MENU_ELEMENTS)  //have we reached the end of the menu?
      {
        u08MenuSelection++; //no, then we can advance the menu
        boolUpdateDisplay=true; //allow the display to update
      }
    }  
    if(u08State==1) //we must not be in the menu area, we are updating the user values
    {
      s16LowerMovement++;   //change the user value, no boundry checking now

      switch (u08MenuSelection) 
      {
        case 0:     //no left and right action here in the user area

        break;
        case 1:     //This is the gaming dim level
          u08DimLevel++;
          if(u08DimLevel>DIM_NUM_ELEMENTS)
          {
            u08DimLevel=0;
          }
          u08DimTarget=pgm_read_byte(&(dimLevel[u08DimLevel]));  
          u16OutputLevel=u08DimTarget;   //dim the PWM           
        break;
        case 2:   //this allows the user to adjust how long the LED's will be on, once they are turned on     
          u08LedSleepLevel++;
          if(u08LedSleepLevel>LED_NUM_ELEMENTS)
          {
            u08LedSleepLevel=0;
          }
          u08LedSleepTarget=pgm_read_byte(&(ledSleepLevel[u08LedSleepLevel]));             
        break;
        case 3:     //This allows the user to adjust when the display goes to sleep
          u08DisplaySleepLevel++;
          if(u08DisplaySleepLevel>DISPLAY_NUM_ELEMENTS)
          {
            u08DisplaySleepLevel=0;
          }
          u08DisplaySleepTarget=pgm_read_byte(&(displaySleepLevel[u08DisplaySleepLevel]));
        break;
        case 4:   //This allows the user to adjust what the controller will do when it starts up (turn on the light full, on dim, or leave it off)
          u08StartUpState++;
          if(u08StartUpState>START_UP_NUM_ELEMENTS)
          {
            u08StartUpState=0;
          }
          //Serial.println(F("u08StartUp e= "));
          //Serial.println(u08StartUpState);
        break;
        case 5:   //This allows to the user to adjust the on board heartbeat indicator
          if(u08Heartbeat!=0)   //wrap around, when the rotary encoder is turned in one direction, go between on-off-on- etc.
          {
            u08Heartbeat=0;
          }
          else
          {
            u08Heartbeat=1;
          }
        break;
      }
      boolUpdateDisplay=true;   //allow the display to update
    }
  }
  else  //we must be turning the rotary encoder the other direction.
  {
    if(u08State==0) //allow the menu to change, since we are in that area
    {
      if(u08MenuSelection>0)    //have we reached the bottom of the menu selection?
      {
        u08MenuSelection--;   //no, then allow the menu to move to the left (down)
        boolUpdateDisplay=true; //allow the display to update
      } 
    }
    if(u08State==1)   //We are in the user value area
    {
      s16LowerMovement--;   //adjust the user value down
      switch (u08MenuSelection)   //the roatary encoder has decrmented a value, we are taking a user value down
      {
        case 0:   //no left and right action here in the user area

        break;
        case 1:   //This is the gaming dim level
          if(u08DimLevel>0)
          {
            u08DimLevel--;  //the rotary encoder knob is being turned left, so decrement the count
          }
          else
          {
            u08DimLevel=DIM_NUM_ELEMENTS; //we have reached the bottom of the dimming, so wrap around to the top
          }
          u08DimTarget=pgm_read_byte(&(dimLevel[u08DimLevel])); //get the actual dim level out of the table
          u16OutputLevel=u08DimTarget;   //load in the dim value to be sent to the OCR1A register, for the PWM
        break;
        case 2: //this allows the user to adjust how long the LED's will be on, once they are turned on               
          if(u08LedSleepLevel>0)
          {
            u08LedSleepLevel--; //decrement the pointer
          }
          else
          {
            u08LedSleepLevel=LED_NUM_ELEMENTS;  //once we are at the bottom of the table, go back to the top
          }
          u08LedSleepTarget=pgm_read_byte(&(ledSleepLevel[u08LedSleepLevel]));             
        break;
        case 3:   //This allows the user to adjust when the display goes to sleep
          if(u08DisplaySleepLevel>0)
          {
            u08DisplaySleepLevel--; //take the display sleep value down
          }
          else
          {
            u08DisplaySleepLevel=DISPLAY_NUM_ELEMENTS;    //we are at the bottom of the table
          }
          u08DisplaySleepTarget=pgm_read_byte(&(displaySleepLevel[u08DisplaySleepLevel]));
        break;
        case 4:   //This allows the user to adjust what the controller will do when it starts up (turn on the light full, on dim, or leave it off)
          if(u08StartUpState>0)
          {
            u08StartUpState--;
          }
          else
          {
            u08StartUpState=START_UP_NUM_ELEMENTS;
          }
        break;
        case 5:   //This allows to the user to adjust the on board heartbeat indicator
          if(u08Heartbeat!=0)   //wrap around, when the rotary encoder is turned in one direction, go between on-off-on- etc.
          {
            u08Heartbeat=0;
          }
          else
          {
            u08Heartbeat=1;
          }
        break;
      }
      boolUpdateDisplay=true; //allow the display to update
    }
  }  
}
